"""
This module provides the `~sunpy.timeseries.TimeSeriesFactory` class.
"""

import os
import copy
import pathlib
from collections import OrderedDict
from urllib.request import Request

import fsspec
import numpy as np
import pandas as pd

import astropy
import astropy.io.fits
import astropy.units as u
from astropy.table import Table
from astropy.time import Time

import sunpy
from sunpy.data import cache
from sunpy.io._file_tools import UnrecognizedFileTypeError, detect_filetype, read_file
from sunpy.io._header import FileHeader
from sunpy.timeseries.sources import source_names
from sunpy.timeseries.timeseriesbase import GenericTimeSeries
from sunpy.util import expand_list
from sunpy.util.datatype_factory_base import (
    BasicRegistrationFactory,
    MultipleMatchError,
    NoMatchError,
    ValidationFunctionError,
)
from sunpy.util.exceptions import warn_user
from sunpy.util.functools import seconddispatch
from sunpy.util.io import HDPair, expand_fsspec_open_file, is_uri, is_url, parse_path, possibly_a_path
from sunpy.util.metadata import MetaDict

__all__ = ["TimeSeries", "TimeSeriesFactory", "NoTimeSeriesFound", "InvalidTimeSeriesInput", "InvalidTimeSeriesType"]


class TimeSeriesFactory(BasicRegistrationFactory):
    """
    A factory for generating solar timeseries objects.

    This factory takes a variety of inputs to generate
    `~sunpy.timeseries.GenericTimeSeries` objects.

    Parameters
    ----------
    \\*inputs
        Inputs to parse for timeseries objects. See the example section for a
        detailed list of possible inputs.

    source : {source_names}, optional
        A string to select the observational source of the data, currently
        necessary to define how files should be read for all instruments.

    concatenate : `bool`, optional
        Defaults to `False`.
        If set, combine any resulting list of TimeSeries objects into a single
        TimeSeries, using successive concatenate methods.

    Returns
    -------
    `sunpy.timeseries.GenericTimeSeries`
        If the input results in a single timeseries object that will be returned, or if ``concatenate=True``.

    `list` of `~sunpy.timeseries.GenericTimeSeries`
        If multiple inputs are parsed, they will be returned in a list, unless
        ``concatenate=True`` is set when they will be combined into a single
        timeseries.

    Examples
    --------
    >>> import sunpy.timeseries
    >>> import sunpy.data.sample  # doctest: +REMOTE_DATA
    >>> my_timeseries = sunpy.timeseries.TimeSeries(sunpy.data.sample.GOES_XRS_TIMESERIES)  # doctest: +REMOTE_DATA

    The SunPy TimeSeries factory accepts a wide variety of inputs for creating timeseries

    * Preloaded tuples of (data, header) pairs or (data, header, units)

    >>> my_timeseries = sunpy.timeseries.TimeSeries((data, header))   # doctest: +SKIP

    Headers and units must be either a `dict`, `~collections.OrderedDict` or `~sunpy.util.metadata.MetaDict`.

    * data, header pairs, or data, header units triples, not in tuples

    >>> my_timeseries = sunpy.timeseries.TimeSeries(data, header)  # doctest: +SKIP
    >>> my_timeseries = sunpy.timeseries.TimeSeries(data, header, units)  # doctest: +SKIP

    * File names for files understood by the file reader and for those that are not

    >>> my_timeseries = sunpy.timeseries.TimeSeries('filename.fits')   # doctest: +SKIP
    >>> my_timeseries = sunpy.timeseries.TimeSeries('filename.fits', source='lyra')  # doctest: +SKIP

    * Multiple files can be combined into one TimeSeries, as long as they are the same source

    >>> my_timeseries = sunpy.timeseries.TimeSeries(['goesfile1.fits', 'goesfile2.fits'],
    ...                                             concatenate=True)  # doctest: +SKIP

    * All fits files in a directory by giving a directory

    >>> my_timeseries = sunpy.timeseries.TimeSeries('local_dir/sub_dir')  # doctest: +SKIP

    * Some regex globs

    >>> my_timeseries = sunpy.timeseries.TimeSeries('eit_*.fits')  # doctest: +SKIP

    * URLs

    >>> my_timeseries = sunpy.timeseries.TimeSeries(url)  # doctest: +SKIP

    * Lists of any of the above

    >>> my_timeseries = sunpy.timeseries.TimeSeries(['file1.fits', 'file2.fits',
    ...                                              'file3.fits', 'directory1/'])  # doctest: +SKIP

    * Any mixture of the above not in a list

    >>> my_timeseries = sunpy.timeseries.TimeSeries((data, header), data2, header2,
    ...                                             'file1.fits', url, 'eit_*.fits')  # doctest: +SKIP
    """

    @staticmethod
    def _read_file(fname, **kwargs):
        """
        Reading a file with automatic source detection.

        Parameters
        ----------
        fname : `str`
            The file path to parse.

        Returns
        -------
        pairs : `list` or `str`
            List of ``(data, header)`` pairs or ``fname`` if the file is not supported or incorrect.
        """
        if "source" not in kwargs.keys() or not kwargs["source"]:
            try:
                if detect_filetype(fname, **kwargs) == "cdf":
                    # Put import here to ensure there is no import dependency
                    # on cdflib for TimeSeries
                    from sunpy.io._cdf import read_cdf

                    return read_cdf(os.fspath(fname), **kwargs)
            except UnrecognizedFileTypeError:
                pass

            try:
                pairs = read_file(os.fspath(fname), **kwargs)

                new_pairs = []
                for pair in pairs:
                    filedata, filemeta = pair
                    if isinstance(filemeta, FileHeader):
                        data = filedata
                        meta = MetaDict(filemeta)
                        new_pairs.append(HDPair(data, meta))
                return [new_pairs]
            except UnrecognizedFileTypeError:
                return [fname]
        else:
            return [fname]

    @staticmethod
    def _is_metadata(meta):
        """
        Return `True` if ``meta`` is an object that could store metadata.
        """
        return isinstance(meta, astropy.io.fits.header.Header | sunpy.io._header.FileHeader | dict | sunpy.timeseries.TimeSeriesMetaData)

    @staticmethod
    def _is_units(units):
        """
        Return `True` if ``units`` is an object that could store units.

        Should be a dictionary of some form (but not `sunpy.util.metadict.MetaDict`)
        with only `astropy.units` for values.
        """
        # Must be a dict and all items must be a unit
        return (isinstance(units, dict)
                and not isinstance(units, MetaDict)
                and all(isinstance(units[key], u.UnitBase) for key in units))

    @staticmethod
    def _from_table(t):
        """
        Extract the data, metadata and units from an astropy table for use in
        constructing a `~sunpy.timeseries.TimeSeries`.

        Parameters
        ----------
        t : `~astropy.table.Table`
            The input table. The datetime column must be the first column or the (single) primary key index.
        """
        table = copy.deepcopy(t)
        # Default the time index to the first column
        index_name = table.colnames[0]
        # Check if another column is defined as the index/primary_key
        if table.primary_key:
            # Check there is only one primary_key/index column
            if len(table.primary_key) == 1:
                table.primary_key[0]
            else:
                raise ValueError("Invalid input Table, TimeSeries doesn't support conversion"
                                 " of tables with more then one index column.")

        # Extract, convert and remove the index column from the input table
        index = table[index_name]
        # Convert if the index is given as an astropy Time object
        if isinstance(index, Time):
            index = index.datetime
        index = pd.to_datetime(index)
        table.remove_column(index_name)

        # Extract the column values from the table
        data = {}
        units = {}
        for colname in table.colnames:
            data[colname] = table[colname]
            units[colname] = table[colname].unit

        # Create a dataframe with this and return
        df = pd.DataFrame(data=data, index=index)
        return df, MetaDict(table.meta), units

    def _parse_meta(self, meta):
        """
        Parse different metadata objects into a MetaDict.
        """
        if isinstance(meta, astropy.io.fits.header.Header):
            meta = MetaDict(sunpy.io._header.FileHeader(meta))
        if isinstance(meta, sunpy.timeseries.TimeSeriesMetaData):
            new_meta = MetaDict()
            for m in meta.metas:
                new_meta.update(m)
            meta = new_meta
        return meta

    def _sanitise_args(self, args, **kwargs):
        """
        Sanitise a list of args so that a single argument corresponds to either:

        - (data, header, units) tuple.
        - path-like `pathlib.Path` (e.g. a filename, directory, glob etc.).
        - `urllib.request.Request`.
        - `GenericTimeSeries`.
        """
        # Account for nested lists of items. Simply outputs a single list of
        # items, nested lists are expanded to element level.
        args = expand_list(args)

        # Sanitise the input so that each 'type' of input corresponds to a different
        # class, so single dispatch can be used later
        parsed_args = []
        skip_n_arg = 0
        for i, arg in enumerate(args):
            if skip_n_arg:
                skip_n_arg -= 1
                continue

            if isinstance(arg, (np.ndarray, Table, pd.DataFrame)):
                # Extract data and metadata
                # The next item is data
                data = args[i]
                meta = MetaDict()
                units = OrderedDict()
                if isinstance(data, Table):
                    # We have an Astropy Table:
                    data, new_meta, new_units = self._from_table(data)
                    units.update(new_units)
                    meta.update(new_meta)
                elif isinstance(data, np.ndarray):
                    # We have a numpy ndarray. We assume the first column is a dt index
                    data = pd.DataFrame(data=data[:, 1:], index=Time(data[:, 0]))

                # The next two could be metadata or units
                for z in range(2):
                    j = i + z + 1
                    if j < len(args):
                        narg = args[j]
                        if self._is_units(narg):
                            units.update(narg)
                            skip_n_arg += 1
                        elif self._is_metadata(narg):
                            meta.update(self._parse_meta(narg))
                            skip_n_arg += 1

                parsed_args.append((data, meta, units))

            elif isinstance(arg, str) and is_url(arg):
                parsed_args.append(Request(arg))
            elif possibly_a_path(arg) and not is_uri(arg):
                parsed_args.append(pathlib.Path(arg))
            elif is_uri(arg):
                fsspec_kw = kwargs.get("fsspec_kwargs", {})
                # Get a list of open file objects (i.e. if given a glob)
                open_files = fsspec.open_files(arg, **fsspec_kw)
                # If any of the elements of the list are directories we want to
                # glob all files in them
                expanded_files = expand_list([expand_fsspec_open_file(f) for f in open_files])
                parsed_args += expanded_files
            else:
                parsed_args.append(arg)
        return parsed_args

    def _parse_args(self, *args, **kwargs):
        """
        Parses an `args` list for data-header pairs. `args` can contain any mixture of the following
        entries:

        * tuples of (data, header, unit) (1)
        * data, header not in a tuple (1)
        * filename, which will be read
        * `pathlib.Path`, which is the filepath to a file to be read.
        * directory, from which all files will be read
        * glob, from which all files will be read
        * url, which will be downloaded and read
        * lists containing any of the above.

        (1) header/unit are optional and in either order, but data should be the first entry in each group.

        Examples
        --------
        self._parse_args(data, header,
                         (data, header),
                         ['file1', 'file2', 'file3'],
                         'file4',
                         'directory1',
                         '*.fits')
        """
        args = self._sanitise_args(args, **kwargs)
        all_ts = []
        for arg in args:
            try:
                all_ts += self._parse_arg(arg, **kwargs)
            except (NoMatchError, MultipleMatchError, ValidationFunctionError) as e:
                msg = f"One of the files failed to validate with: {e}"
                if self.allow_errors:
                    warn_user(msg)
                    continue
                else:
                    msg += "\nTo bypass these errors, set `allow_errors=True`."
                    raise type(e)(msg) from e
            except Exception as e:
                msg = f"Something went wrong: {e}"
                raise type(e)(msg) from e
        return all_ts

    def _parse_ts_results(self, r, **kwargs):
        """
        Prepare parsed result from into correct format for _parse_args.
        r can be TimeSeries, path, or a data, header pair
        **kwargs get passed to `_get_matching_widget` and `_check_registered_widgets`
        """
        if isinstance(r, GenericTimeSeries):
            return [r]
        elif isinstance(r, pathlib.Path):
            return [self._check_registered_widgets(filepath=r, **kwargs)]
        else:
            pairs = r
            # Pairs may be x long where x is the number of HDUs in the file.
            headers = [pair.header for pair in pairs]

            types = []
            for header in headers:
                try:
                    match = self._get_matching_widget(meta=header, **kwargs)
                    if not match == GenericTimeSeries:
                        types.append(match)
                except (MultipleMatchError, NoMatchError):
                    continue

            if not types:
                # If no specific classes have been found we can read the data
                # if we only have one data header pair:
                if len(pairs) == 1:
                    return [GenericTimeSeries(pairs[0]._data, pairs[0].header)]

                else:
                    raise NoMatchError(
                        "Input read by sunpy.io can not find a " "matching class for reading multiple HDUs"
                    )
            if len(set(types)) > 1:
                raise MultipleMatchError("Multiple HDUs return multiple matching classes.")

            cls = types[0]

            data_header_unit_tuple = cls._parse_hdus(pairs)
            return self._parse_arg(data_header_unit_tuple)

    @seconddispatch
    def _parse_arg(self, arg, **kwargs):
        """
        Parse a single arg and return a list of timeseries.
        """
        raise NoMatchError(f"File not found or invalid input with {arg} and {kwargs}")

    @_parse_arg.register(GenericTimeSeries)
    def _parse_ts(self, ts, **kwargs):
        return [ts]

    @_parse_arg.register(Request)
    def _parse_url(self, request, **kwargs):
        path = cache.download(request.full_url).absolute()
        return self._parse_path(pathlib.Path(path), **kwargs)

    @_parse_arg.register(pathlib.Path)
    def _parse_path(self, path, **kwargs):
        results = parse_path(path, self._read_file, **kwargs)
        all_ts = []
        for r in results:
            all_ts += self._parse_ts_results(r, **kwargs)
        return all_ts

    @_parse_arg.register(tuple)
    def _parse_tuple(self, tup, **kwargs):
        data, header, units = tup
        # Make a MetaDict from various input types
        meta = header
        if isinstance(meta, astropy.io.fits.header.Header):
            meta = sunpy.io._header.FileHeader(meta)
        meta = MetaDict(meta)
        return [self._check_registered_widgets(data=data, meta=meta, units=units, **kwargs)]

    @_parse_arg.register(fsspec.core.OpenFile)
    def _parse_fsspec_file(self, arg, **kwargs):
        # TODO: We should probably migrate the whole of the unified IO layer to
        # use fsspec for everything, but for now we parse the URI through
        return self._read_file(arg.full_name, **kwargs)

    def __call__(self, *args, allow_errors=False, **kwargs):
        """
        Method for running the factory. Takes arbitrary arguments and keyword
        arguments and passes them to a sequence of pre-registered types to
        determine which is the correct `~sunpy.timeseries.TimeSeries` source
        type to build.

        Arguments args and kwargs are passed through to the validation function and to the constructor for the final type.
        For `~sunpy.timeseries.TimeSeries` types, validation function must take a data-header pair as an argument.

        Parameters
        ----------
        allow_errors : `bool`, optional
            If set, bypass data-header pairs or files which cause an exception and warn instead.
            Defaults to `False`.

        Notes
        -----
        Extra keyword arguments are passed through to `sunpy.io.read_file` such as `memmap` for FITS files.
        """
        self.allow_errors = allow_errors
        new_timeseries = self._parse_args(*args, **kwargs)

        # Concatenate the timeseries into one if specified.
        concatenate = kwargs.get("concatenate", False)
        if concatenate:
            # Merge all these timeseries into one.
            full_timeseries = new_timeseries.pop(0)
            for timeseries in new_timeseries:
                full_timeseries = full_timeseries.concatenate(timeseries)

            new_timeseries = [full_timeseries]

        # Only return single time series, not in a list if we only have one.
        if len(new_timeseries) == 1:
            return new_timeseries[0]
        return new_timeseries

    def _get_matching_widget(self, **kwargs):
        candidate_widget_types = list()

        for key in self.registry:
            # Call the registered validation function for each registered class
            if self.registry[key](**kwargs):
                candidate_widget_types.append(key)

        n_matches = len(candidate_widget_types)

        if n_matches == 0:
            if self.default_widget_type is None:
                raise NoMatchError("No types match specified arguments and no default is set.")
            else:
                candidate_widget_types = [self.default_widget_type]
        elif n_matches > 1:
            raise MultipleMatchError(
                f"Too many candidate types identified ({n_matches})."
                "Specify enough keywords to guarantee unique type "
                "identification."
            )

        # Only one suitable source class is found
        return candidate_widget_types[0]

    def _check_registered_widgets(self, **kwargs):
        """
        Checks the (instrument) source(s) that are compatible with this given
        file/data.

        Only if exactly one source is compatible will a
        `~sunpy.timeseries.TimeSeries` be returned.
        """
        WidgetType = self._get_matching_widget(**kwargs)

        # Dealing with the fact that timeseries filetypes are less consistent
        # (then maps), we use a _parse_file() method embedded into each
        # instrument subclass.
        filepath = kwargs.pop("filepath", None)
        data = kwargs.pop("data", None)
        meta = kwargs.pop("meta", None)
        units = kwargs.pop("units", None)
        if filepath:
            data, meta, units = WidgetType._parse_file(filepath)

        # Now return a TimeSeries from the given file.
        return WidgetType(data, meta, units, **kwargs)


def _apply_result(data_header_pairs, filepaths, result):
    read, result = result
    if read:
        data_header_pairs.append(result)
    else:
        filepaths.append(result)

    return data_header_pairs, filepaths


class InvalidTimeSeriesInput(ValueError):
    """
    Exception to raise when input variable is not a
    `~sunpy.timeseries.TimeSeries` instance and does not point to a valid
    TimeSeries input file.
    """


class InvalidTimeSeriesType(ValueError):
    """
    Exception to raise when an invalid type of timeseries is requested with
    `~sunpy.timeseries.TimeSeries`.
    """


class NoTimeSeriesFound(ValueError):
    """
    Exception to raise when input does not point to any valid
    `~sunpy.timeseries.TimeSeries` or files.
    """


TimeSeriesFactory.__doc__ = TimeSeriesFactory.__doc__.format(source_names=source_names)
TimeSeries = TimeSeriesFactory(registry=GenericTimeSeries._registry,
                               default_widget_type=GenericTimeSeries,
                               additional_validation_functions=["is_datasource_for"])



"""
===========================
Adding an Earth scale image
===========================

This example shows how to plot a map with an image of the Earth added for scale.
"""
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image

import astropy.units as u
from astropy.constants import R_earth
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS

import sunpy.map
from sunpy.coordinates.utils import solar_angle_equivalency
from sunpy.data import EARTH_IMAGE
from sunpy.data.sample import AIA_193_CUTOUT01_IMAGE

###############################################################################
# We start with a sample AIA image.

cutout_map = sunpy.map.Map(AIA_193_CUTOUT01_IMAGE)

###############################################################################
# We use a (low-resolution) image of Earth that we provide with `sunpy`. You
# can use a different image as desired, and it should have a transparent
# background. We also crop the image tightly so that we can assume that the
# image width/height are equal to the Earth diameter. Finally, we flip the
# image vertically so that it is oriented correctly when plotted with the pixel
# origin in the lower-left corner, which is the convention for maps.

earth = Image.open(EARTH_IMAGE)
earth = earth.crop(earth.getbbox())
earth = earth.transpose(Image.FLIP_TOP_BOTTOM)

##############################################################################
# The first step in plotting the Earth is to convert the Earth's diameter in km
# to arcsec on the plane-of-sky for the time of the observation.  We use the
# observer coordinate from the map. For more information about the
# transformation, see :func:`~sunpy.coordinates.utils.solar_angle_equivalency`.

earth_diameter = 2 * R_earth.to(u.arcsec, equivalencies=solar_angle_equivalency(cutout_map.observer_coordinate))
print(f"Earth diameter = {earth_diameter:.2f}")

##############################################################################
# We then create a WCS header so that the Earth will be plotted correctly
# on any underlying map projection. We center the Earth on a specific point in
# the map.

earth_x = 1000 * u.arcsec
earth_y = -200 * u.arcsec
earth_center = SkyCoord(earth_x, earth_y, frame=cutout_map.coordinate_frame)
earth_wcs = sunpy.map.make_fitswcs_header(
    (earth.height, earth.width), earth_center,
    scale=earth_diameter / ([earth.width, earth.height] * u.pix)
)

##############################################################################
# Now we plot the AIA image and superimpose the Earth for scale. We use
# :meth:`matplotlib.axes.Axes.pcolormesh` to plot the pixels of the Earth
# image as a quadrilateral mesh, which accommodates any WCS of the underlying
# map. We also add a text label above the Earth.

fig = plt.figure()
ax = fig.add_subplot(projection=cutout_map)
cutout_map.plot(clip_interval=(1, 99.9)*u.percent)

ax.pcolormesh(
    *np.meshgrid(np.arange(earth.width + 1) - 0.5, np.arange(earth.height + 1) - 0.5),
    earth, shading='flat', transform=ax.get_transform(WCS(earth_wcs))
)

ax.text(
    earth_x.to_value('deg'), (earth_y + earth_diameter).to_value('deg'),
    'Earth to scale', color='white', fontsize=12, horizontalalignment='center',
    transform=ax.get_transform('world')
)

plt.show()



"""
This module provides the following sample data files. When a sample shortname
is accessed, the corresponding file is downloaded if needed. All files can be
downloaded by calling :func:`~sunpy.data.sample.download_all`.

Summary variables
-----------------
.. list-table::
   :widths: auto

   * - ``file_dict``
     - Dictionary of all sample shortnames and, if downloaded, corresponding
       file locations on disk (otherwise, ``None``)
   * - ``file_list``
     - List of disk locations for sample data files that have been downloaded

Sample shortnames
-----------------
.. list-table::
   :widths: auto
   :header-rows: 1

   * - Sample shortname
     - Name of downloaded file
"""
from ._sample import _SAMPLE_DATA, _get_sample_files

# Add a table row to the module docstring for each sample file
for _keyname, _filename in sorted(_SAMPLE_DATA.items()):
    __doc__ += f'   * - ``{_keyname}``\n     - {_filename}\n'


# file_dict and file_list are not normal variables; see __getattr__() below
__all__ = list(sorted(_SAMPLE_DATA.keys())) + ['download_all', 'file_dict', 'file_list']  # NOQA: F822


# See PEP 562 (https://peps.python.org/pep-0562/) for module-level __dir__()
def __dir__():
    return __all__


# See PEP 562 (https://peps.python.org/pep-0562/) for module-level __getattr__()
def __getattr__(name):
    if name in _SAMPLE_DATA:
        return _get_sample_files([_SAMPLE_DATA[name]])[0]
    elif name == 'file_dict':
        return dict(sorted(zip(_SAMPLE_DATA.keys(),
                               _get_sample_files(_SAMPLE_DATA.values(), no_download=True))))
    elif name == 'file_list':
        return [v for v in __getattr__('file_dict').values() if v]
    else:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


def download_all(force_download=False):
    """
    Download all sample data at once that has not already been downloaded.

    Parameters
    ----------
    force_download : `bool`
        If ``True``, files are downloaded even if they already exist. Default is
        ``False``.
    """
    _get_sample_files(_SAMPLE_DATA.values(), force_download=force_download)



"""
Test cases for STEREO CORMap subclass.
"""
import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.stereo import CORMap
from .helpers import _test_private_date_setters


@pytest.fixture
def cor_map():
    header_file = get_test_filepath("cor1_20090615_000500_s4c1A.header")
    return get_dummy_map_from_header(header_file)


def test_fits_to_cor(cor_map):
    """Tests the creation of CORMap using FITS."""
    assert isinstance(cor_map, CORMap)


def test_reference_date(cor_map):
    assert cor_map.reference_date.isot == "2009-06-15T00:05:00.855"


def test_date(cor_map):
    assert cor_map.reference_date.isot == "2009-06-15T00:05:00.855"


def test_private_date_setters(cor_map):
    _test_private_date_setters(cor_map)


def test_is_datasource_for(cor_map):
    """Test the is_datasource_for method of CORMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert cor_map.is_datasource_for(cor_map.data, cor_map.meta)


def test_measurement(cor_map):
    """Tests the measurement property of the CORMap object."""
    assert cor_map.measurement == "white-light"


def test_observatory(cor_map):
    """Tests the observatory property of the CORMap object."""
    assert cor_map.observatory == "STEREO A"


def test_norm_clip(cor_map):
    # Tests that the default normalizer has clipping disabled
    assert not cor_map.plot_settings['norm'].clip


def test_wcs(cor_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    cor_map.pixel_to_world(0*u.pix, 0*u.pix)



"""
The following colormaps are provided by this module.

.. plot::

    import matplotlib.pyplot as plt
    import sunpy.visualization.colormaps as cm
    cm.show_colormaps()


"""
from sunpy.visualization.colormaps.cm import *

for cmname in cmlist.keys():  # NOQA: F405
    __doc__ += f"\n* '{cmname}'\n"



import copy

import numpy as np
import pytest

import astropy.units as u
import astropy.wcs
from astropy.io import fits
from astropy.io.fits.verify import VerifyWarning
from astropy.wcs import FITSFixedWarning

import sunpy.map
from sunpy.data.test import get_test_filepath


@pytest.fixture
def aia171_test_dask_map(aia171_test_map):
    dask_array = pytest.importorskip('dask.array')
    return aia171_test_map._new_instance(
        dask_array.from_array(aia171_test_map.data),
        copy.deepcopy(aia171_test_map.meta)
    )


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_dask_array_repr(aia171_test_dask_map):
    # Check that _repr_html_ functions for a dask array
    html_dask_repr = aia171_test_dask_map._repr_html_(compute_dask=False)
    html_computed_repr = aia171_test_dask_map._repr_html_(compute_dask=True)
    assert html_dask_repr != html_computed_repr


# This is needed for the reproject_to function
with pytest.warns(VerifyWarning, match="Invalid 'BLANK' keyword in header."):  # NOQA: PT031
    with fits.open(get_test_filepath('aia_171_level1.fits')) as hdu:
        with pytest.warns(FITSFixedWarning, match="'datfix' made the change"):
            aia_wcs = astropy.wcs.WCS(header=hdu[0].header)


@pytest.mark.parametrize(("func", "args"), [
    ("max", {}),
    ("mean", {}),
    ("min", {}),
    pytest.param("reproject_to", {"wcs": aia_wcs}, marks=pytest.mark.xfail(reason="reproject is not dask aware")),
    pytest.param("resample", {"dimensions": (100, 100)*u.pix}, marks=pytest.mark.xfail()),
    pytest.param("rotate", {}, marks=pytest.mark.xfail(reason="nanmedian is not implemented in Dask")),
    ("std", {}),
    ("superpixel", {"dimensions": (10, 10)*u.pix}),
    ("submap", {"bottom_left": (100, 100)*u.pixel, "width": 10*u.pixel, "height": 10*u.pixel}),
])
def test_method_preserves_dask_array(aia171_test_map, aia171_test_dask_map, func, args):
    """
    Check that map methods preserve dask arrays if they are given as input, instead of eagerly
    operating on them and bringing them into memory.
    """
    # Check that result array is still a Dask array
    res_dask = aia171_test_dask_map.__getattribute__(func)(**args)
    res = aia171_test_map.__getattribute__(func)(**args)
    result_is_map = isinstance(res, sunpy.map.GenericMap)
    if result_is_map:
        assert isinstance(res_dask.data, type(aia171_test_dask_map.data))
        assert np.allclose(res_dask.data.compute(), res.data, atol=0.0, rtol=0.0)
    else:
        assert isinstance(res_dask, type(aia171_test_dask_map.data))
        assert np.allclose(res_dask.compute(), res, atol=0.0, rtol=0.0)




import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.adapt import ADAPTMap
from sunpy.util.exceptions import SunpyMetadataWarning
from .helpers import _test_private_date_setters


@pytest.fixture
def adapt_map():
    header_file = get_test_filepath("adapt.header")
    return get_dummy_map_from_header(header_file)


def test_date(adapt_map):
    """Tests the date property of the ADAPTMap object."""
    assert adapt_map.date.isot == "2020-01-01T00:00:00.000"


def test_reference_date(adapt_map):
    """Tests the reference_date property of the ADAPTMap object."""
    assert adapt_map.reference_date.isot == "2020-01-01T00:00:00.000"


def test_private_date_setters(adapt_map):
    _test_private_date_setters(adapt_map)


def test_coordinate_system(adapt_map):
    assert adapt_map.coordinate_system == ('CRLN-CAR', 'CRLT-CAR')


def test_fitstoadapt(adapt_map):
    """Tests the creation of ADAPTMap using FITS."""
    assert isinstance(adapt_map, ADAPTMap)
    assert adapt_map.data.shape == (180, 360)


def test_is_datasource_for(adapt_map):
    """Test the is_datasource_for method of ADAPTMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert adapt_map.is_datasource_for(adapt_map.data, adapt_map.meta)


def test_measurement(adapt_map):
    """Tests the measurement property of the ADAPTMap object."""
    assert adapt_map.measurement is None


def test_observatory(adapt_map):
    """Tests the observatory property of the ADAPTMap object."""
    assert adapt_map.observatory == ""


def test_wcs(adapt_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    with pytest.warns(SunpyMetadataWarning, match='assuming Earth-based observer'):
        adapt_map.pixel_to_world(0*u.pix, 0*u.pix)



from inspect import cleandoc

import numpy as np
import pytest

from astropy.io import fits

from sunpy.data.test import get_test_filepath
from sunpy.util import util
from sunpy.util.parfive_helpers import Results


def test_unique():
    """
    This should add the unique values of itr to unique_list.
    """
    itr = [6, 1, 2, 1, 7, 41.2, '41.2', 1, '41.2']
    unique_list = []
    for elem in util.unique(itr):
        unique_list.append(elem)
    assert unique_list == [6, 1, 2, 7, 41.2, '41.2']


def test_unique_key():
    """
    This should add each element of itr to unique_list if no preceding element
    is congruent to it in mod 10.
    """
    itr = [7, 3, 17, 104, 6, 1006, 117, 14, 10]
    unique_list = []
    for elem in util.unique(itr, lambda x: x % 10):
        unique_list.append(elem)
    assert unique_list == [7, 3, 104, 6, 10]


def test_replacement_filename():
    """
    This should return a replacement path for the current file.
    """
    assert util.replacement_filename(__file__).endswith('test_util.0.py')


def test_replacement_filename_path_not_exists(mocker):
    """
    If a candidate path does not exist, then just return it as it is OK to use.
    """
    path_not_exists = '/tmp'
    mocker.patch('os.path.exists', return_value=False)

    assert util.replacement_filename(path_not_exists) == path_not_exists


def test_expand_list():
    """
    This should return an expanded version of list lst.
    """
    lst = [1, 2, 3, [4, 5, 6], 7, (8, 9), ((10, 11), ((12, 13),))]
    assert util.expand_list(lst) == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]


@pytest.mark.parametrize(('input_data', 'expected_output'), [
    (['a1234', 'b', [], (['c', 'd']), (), ['e'], b'fghj'], ['a1234', 'b', 'c', 'd', 'e', b'fghj']),
    (iter(['c', 'd']), ['c', 'd']),
    (zip(['1', '2'], ['3', '4']), ['1', '3', '2', '4']),
    (Results(["a", "b", "c"]), ["a", "b", "c"]),
    ([open(get_test_filepath('aia_171_level1.fits'), 'rb')], open(get_test_filepath('aia_171_level1.fits'), 'rb').readlines())
])
def test_expand_list_generator(input_data, expected_output):
    assert list(util.expand_list_generator(input_data)) == expected_output


@pytest.mark.filterwarnings("ignore:Invalid 'BLANK' keyword in header")
def test_expand_list_generator_map(aia171_test_map):
    before = after = [aia171_test_map.data, aia171_test_map.wcs, fits.getheader(get_test_filepath('aia_171_level1.fits'))]
    assert list(util.expand_list_generator(before)) == after


def test_partial_key_match():
    test_dict = {('a', 'b', 'c'): (1, 2, 3), ('a', 'b', 'd'): (4, 5, 6), ('e', 'f', 'g'): (8, 7, 9)}
    assert list(util.partial_key_match(('a', None, 'c'), test_dict))[
        0] == test_dict[('a', 'b', 'c')]


def test_dict_keys_same():
    dicts = [{'x': 42}, {'x': 23, 'y': 5}]
    assert util.dict_keys_same(dicts) == [{'y': None, 'x': 42}, {'y': 5, 'x': 23}]


def test_get_keywords():

    def f(a, b, c=1, d=2, **e):
        pass

    def g(a, b, *, c=1, h=2, **e):
        pass

    assert util.get_keywords(f) == {'c', 'd'}  # POSITIONAL_OR_KEYWORD
    assert util.get_keywords(g) == {'c', 'h'}  # KEYWORD_ONLY
    assert util.get_keywords([f, g]) == {'c', 'd', 'h'}


def test_get_set_methods():

    class A:

        def _set_test1(self, *args, **kwargs):
            pass

        def set_test2(self, *args, **kwargs):
            pass

        @property
        def set_test3(self):
            pass

        def test4(self, *args, **kwargs):
            pass

        @property
        def test5(self, *args, **kwargs):
            pass

        def set_test6_n(self, *args, **kwargs):
            pass

    assert util.get_set_methods(A()) == {'test2', 'test6_n'}


def test_notes_combined():
    original_documentation = """
    Class Info.

    Notes
    -----
    This is a note.

    References
    ----------
    This is reference.
    """
    extra_note_section = """\nNotes\n-----\nThis should be combined."""
    updated_documentation = util.fix_duplicate_notes(extra_note_section, original_documentation)
    expected_result = """
    Class Info.

    Notes
    -----
    This is a note.

    This should be combined.

    References
    ----------
    This is reference.
    """
    assert updated_documentation == cleandoc(expected_result)


def test_notes_combined_no_references():
    original_documentation = """
    Class Info.

    Notes
    -----
    This is a note.
    """
    extra_note_section = """\nNotes\n-----\nThis should be combined."""
    updated_documentation = util.fix_duplicate_notes(extra_note_section, original_documentation)
    expected_result = """
    Class Info.

    Notes
    -----
    This is a note.

    This should be combined.
    """
    assert updated_documentation == cleandoc(expected_result)


def test_notes_combined_no_existing_notes():
    original_documentation = """
    Class Info.

    References
    ----------
    This is reference.
    """
    extra_note_section= """\nNotes\n-----\nThis should be combined."""
    updated_documentation = util.fix_duplicate_notes(extra_note_section, original_documentation)
    expected_result = """
    Class Info.

    Notes
    -----
    This should be combined.

    References
    ----------
    This is reference.
    """
    assert updated_documentation == cleandoc(expected_result)


def test_notes_combined_no_notes_no_references():
    original_documentation = """
    Class Info.
    """
    extra_note_section= """\nNotes\n-----\nThis should be combined."""
    updated_documentation= util.fix_duplicate_notes(extra_note_section, original_documentation)
    expected_result = """
    Class Info.

    Notes
    -----
    This should be combined.
    """
    assert updated_documentation == cleandoc(expected_result)


def test_grid_perimeter():
    assert np.all(util.grid_perimeter(2, 3) == [[0, 0], [0, 1], [0, 2], [0, 3], [1, 3], [2, 3], [2, 2], [2, 1], [2, 0], [1, 0]])




from sunpy.util.metadata import MetaDict

# ==============================================================================
# Test, read, get_header and write through the file independent layer
# ==============================================================================


def test_upcasing():
    meta = MetaDict({'wibble': 1, 'WOBBLE': 2})
    # __getitem__
    assert meta['wibble'] == meta['WIBBLE']
    # get
    assert meta.get('wibble') == meta.get('WIBBLE')
    # has_key
    assert ('wibble' in meta) == ('WIBBLE' in meta)
    # Copy
    meta2 = meta.copy()
    assert meta2 == meta
    # pop
    assert meta.pop('wibble') == meta2.pop('WIBBLE')
    # update
    meta.update({'spam': 'eggs'})
    meta2.update({'SPAM': 'eggs'})
    assert meta == meta2
    # setdefault
    meta.setdefault('dave', 3)
    meta2.setdefault('DAVE', 3)
    assert meta.get('DAVE') == meta2.get('dave')
    # __setitem__
    meta['wibble'] = 10
    assert meta['wibble'] == 10
    meta['WIBBLE'] = 20
    assert meta['wibble'] == 20
    # __contains__
    assert 'wibble' in meta
    assert 'WIBBLE' in meta



import pytest

from sunpy.util.datatype_factory_base import (
    BasicRegistrationFactory,
    MultipleMatchError,
    NoMatchError,
    ValidationFunctionError,
)


class BaseWidget:
    def __init__(self, *args, **kwargs):
        pass


class DefaultWidget(BaseWidget):
    pass


class StandardWidget(BaseWidget):
    @classmethod
    def _factory_validation_function(cls, *args, **kwargs):
        return kwargs.get('style') == 'standard'


class DuplicateStandardWidget(BaseWidget):
    @classmethod
    def _factory_validation_function(cls, *args, **kwargs):
        return kwargs.get('style') == 'standard'


class FancyWidget(BaseWidget):
    @classmethod
    def _factory_validation_function(cls, *args, **kwargs):
        return kwargs.get('style') == 'fancy' and 'feature' in kwargs


class ExternallyValidatedWidget(BaseWidget):
    pass


def external_validation_function(*args, **kwargs):
    return kwargs.get('style') == 'external'


class UnvalidatedWidget(BaseWidget):
    pass


class MissingClassMethodWidget(BaseWidget):
    def _factory_validation_function(cls, *args, **kwargs):
        return kwargs.get('style') == 'missing'


class DifferentValidationWidget(BaseWidget):
    @classmethod
    def different_validation_function(cls, *args, **kwargs):
        return kwargs.get('style') == 'different'


class MissingClassMethodDifferentValidationWidget(BaseWidget):
    def different_validation_function(cls, *args, **kwargs):
        return kwargs.get('style') == 'missing-different'


def test_default_factory():
    DefaultFactory = BasicRegistrationFactory()

    DefaultFactory.register(DefaultWidget, is_default=True)
    assert DefaultFactory.default_widget_type == DefaultWidget

    DefaultFactory.register(StandardWidget)
    DefaultFactory.register(FancyWidget)
    DefaultFactory.register(ExternallyValidatedWidget,
                            validation_function=external_validation_function)

    assert isinstance(DefaultFactory(), DefaultWidget)
    assert isinstance(DefaultFactory(style='standard'), StandardWidget)
    assert isinstance(DefaultFactory(style='fancy'), DefaultWidget)
    assert isinstance(DefaultFactory(style='fancy', feature="present"), FancyWidget)
    assert isinstance(DefaultFactory(style='external'), ExternallyValidatedWidget)

    with pytest.raises(ValidationFunctionError):
        DefaultFactory.register(UnvalidatedWidget)

    with pytest.raises(ValidationFunctionError):
        DefaultFactory.register(MissingClassMethodWidget)

    DefaultFactory.unregister(StandardWidget)
    assert not isinstance(DefaultFactory(style='standard'), StandardWidget)


def test_validation_fun_not_callable():
    TestFactory = BasicRegistrationFactory()

    with pytest.raises(AttributeError):
        TestFactory.register(StandardWidget, validation_function='not_callable')


def test_no_default_factory():
    NoDefaultFactory = BasicRegistrationFactory()
    NoDefaultFactory.register(StandardWidget)
    NoDefaultFactory.register(FancyWidget)

    with pytest.raises(NoMatchError):
        NoDefaultFactory()

    # Raises because all requirements are not met for FancyWidget and no
    # default is present.
    with pytest.raises(NoMatchError):
        NoDefaultFactory(style='fancy')

    assert isinstance(NoDefaultFactory(style='standard'), StandardWidget)
    assert isinstance(NoDefaultFactory(style='fancy', feature='present'), FancyWidget)


def test_with_external_registry():
    external_registry = {}

    FactoryWithExternalRegistry = BasicRegistrationFactory(registry=external_registry)
    assert len(external_registry) == 0

    FactoryWithExternalRegistry.register(StandardWidget)
    assert isinstance(FactoryWithExternalRegistry(style='standard'), StandardWidget)

    # Ensure the 'external_registry' is being populated see #1988
    assert len(external_registry) == 1


def test_multiple_match_factory():
    MultipleMatchFactory = BasicRegistrationFactory()
    MultipleMatchFactory.register(StandardWidget)
    MultipleMatchFactory.register(DuplicateStandardWidget)

    with pytest.raises(MultipleMatchError):
        MultipleMatchFactory(style='standard')


def test_extra_validation_factory():
    ExtraValidationFactory = BasicRegistrationFactory(
        additional_validation_functions=['different_validation_function'])
    ExtraValidationFactory.register(DifferentValidationWidget)
    assert isinstance(ExtraValidationFactory(style='different'), DifferentValidationWidget)

    with pytest.raises(ValidationFunctionError):
        ExtraValidationFactory.register(MissingClassMethodDifferentValidationWidget)



from sunpy.net.attr import SimpleAttr

__all__ = ['ADAPTFileType', 'ADAPTLonType', 'ADAPTInputSource',
           'ADAPTDataAssimilation', 'ADAPTResolution', 'ADAPTVersionYear', 'ADAPTVersionMonth',
           'ADAPTEvolutionMode', 'ADAPTHelioData', 'ADAPTRealizations', 'ADAPTMagData']

# Define a custom __dir__ to restrict tab-completion to __all__
def __dir__():
    return __all__


class ADAPTFileType(SimpleAttr):
    """
    ADAPT file type: Public.
    """
    pass

class ADAPTLonType(SimpleAttr):
    """
    ADAPT longitude type: Carrington Fixed, Central Meridian, East Limb.
    """
    pass

class ADAPTInputSource(SimpleAttr):
    """
    ADAPT input source: All, KPVT, VSM, GONG, HMI, MDI, MWO.
    """
    pass

class ADAPTDataAssimilation(SimpleAttr):
    """
    ADAPT data assimilation: WH, enLS, enkf, enLAKF.
    """
    pass

class ADAPTResolution(SimpleAttr):
    """
    ADAPT model spatial resolution: 1.0 deg, 0.2 deg.
    """
    pass

class ADAPTVersionYear(SimpleAttr):
    """
    ADAPT code version year.
    """
    pass

class ADAPTVersionMonth(SimpleAttr):
    """
    ADAPT code version month.
    """
    pass

class ADAPTEvolutionMode(SimpleAttr):
    """
    ADAPT evolution mode: Data assimilation step, Intermediate step, Forecast step.
    """
    pass

class ADAPTHelioData(SimpleAttr):
    """
    ADAPT helioseismic data: Not added or no data, Far-side, Emergence, Both emergence & far-side.
    """
    pass

class ADAPTRealizations(SimpleAttr):
    """
    ADAPT realizations: number of model/file realizations, e.g., 16 -> "016"
    """
    pass


class ADAPTMagData(SimpleAttr):
    """
    ADAPT magnetic data: Not added or no data, Mag-los, Mag-vector, Mag- both los & vector, Mag- polar avg obs, Mag- los & polar, Mag- vector & polar, Mag- both los and vector & polar.
    """
    pass



"""
This subpackage contains a series of file readers and only for internal use.

None of this code is intended to be used directly by a user and is not part of the sunpy public API
and is not added to the API reference pages.
"""



"""
=================================
Finding and masking bright pixels
=================================

How to find and overplot the location of the brightest
pixel and then mask pixels around that region.
"""
# sphinx_gallery_thumbnail_number = 2

import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma

import astropy.units as u

import sunpy.map
from sunpy.data.sample import AIA_171_IMAGE
from sunpy.map.maputils import all_coordinates_from_map

###############################################################################
# We start with the sample data.

aia = sunpy.map.Map(AIA_171_IMAGE)

###############################################################################
# To find the brightest pixel, we find the maximum in the AIA image data
# then transform that pixel coordinate to a map coordinate.

pixel_pos = np.argwhere(aia.data == aia.data.max()) * u.pixel
hpc_max = aia.wcs.pixel_to_world(pixel_pos[:, 1], pixel_pos[:, 0])

###############################################################################
# Let's plot the results.

fig = plt.figure()
ax = fig.add_subplot(projection=aia)
aia.plot(axes=ax)
ax.plot_coord(hpc_max, color='white', marker='x', markersize=15)
plt.show()

###############################################################################
# A utility function gives us access to the helioprojective coordinate of each
# pixels. We create a new array which contains the normalized radial position
# for each pixel adjusted for the position of the brightest pixel
# (using ``hpc_max``) and then create a new map.

hpc_coords = all_coordinates_from_map(aia)
r_mask = np.sqrt((hpc_coords.Tx - hpc_max.Tx) ** 2 +
                 (hpc_coords.Ty - hpc_max.Ty) ** 2) / aia.rsun_obs
mask = ma.masked_less_equal(r_mask, 0.1)
scaled_map = sunpy.map.Map(aia.data, aia.meta, mask=mask.mask)

###############################################################################
# Let's plot the results.

fig = plt.figure()
ax = fig.add_subplot(projection=scaled_map)
scaled_map.plot(axes=ax)
plt.show()
